!+
! Module: embadpixfind_mod
!
! Authors:  Jean Ballet - SAp   CEA Saclay
!          
! Description: Find EPIC-MOS bad pixels
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module embadpixfind_mod

  use emutils_mod  
  use cal
  use dal
  use param  
  use errorHandling

  implicit none 
  
  private 

! Names of the task parameters (2nd string for the keywords)

    character(len=*), parameter :: outimage    = "evimageset", &
                                   medimage    = "medianset", &
                                   outbadpix   = "badpixset", &
                                   incbadpix   = "incremental", &
                                   threshold   = "probathreshold", &
                                   badcols     = "findbadsegments", &
                                   onbright    = "findbright", &
                                   ondead      = "finddead", &
                                   maxrat      = "maxratio",  &
                                   usecal      = "usecal", &
                                   adddead     = "includedeadpixels"

! Define box width for averages

    integer, parameter      :: w2 = 2, width = 2*w2+1

! Maximum number of bad pixels per CCD

    integer,parameter :: maxbad = 32768

! Keywords to check for compatibility of image files

    character(len=8), dimension(6), parameter :: keycompat = &
         (/"TELESCOP","INSTRUME","CCDID   ","CCDNODE ","OBS_ID  ","EXPIDSTR"/)

! Window parameters

    integer :: wx0, wx1, wy0, wy1

    interface medloc
       module procedure intmedloc
       module procedure realmedloc
    end interface

  public :: embadpixfind

contains 
  
  !---------------------------------------------------------------------
  subroutine embadpixfind
  !---------------------------------------------------------------------

    type(ArrayT) :: ev_map, med_map
    type(TableT) :: bad_tab
    integer(kind=int16), dimension(maxbad) :: xbad, ybad, yext, tbad, fbad
    integer :: nbad
    real    :: maxproba, maxratio
    logical :: withcal, withdead, findcols, finddead, findbright, incremental

! Read parameters

    call getParamValues(ev_map, med_map, bad_tab, incremental, maxproba, &
                   maxratio, findcols, findbright, finddead, withcal, withdead)

    call readBadpix(bad_tab, incremental, xbad, ybad, yext, tbad, fbad, nbad)

! New bad pixels are not uplinked, neither are dead pixels

    fbad(nbad+1:maxbad) = 2

! Look for bad pixels

    call findAllBad(ev_map, med_map, maxproba, maxratio, findcols, findbright,&
               finddead, withcal, withdead, xbad, ybad, yext, tbad, nbad)

! Merge bad pixels aligned along a column

    if (nbad > 1)  call mergeBad(xbad, ybad, yext, tbad, fbad, nbad)

! Write bad pixels file

    call writeBadpix(bad_tab, xbad, ybad, yext, tbad, fbad, nbad)

    call release(parent(ev_map))
    call release(parent(med_map))

  end subroutine embadpixfind

! Read data and call all specific bad structure finders

  subroutine findAllBad(ev_map, med_map, maxproba, maxratio, findcols,  &
         findbright, finddead, withcal, withdead, xbad, ybad, yext, tbad, nbad)

! ev_map     : Image array
! med_map    : Same image after median smoothing
! maxproba   : Probability threshold for false detection
! maxratio   : Relative detection threshold for dark features
! findcols   : Look for bad rows or columns if true
! findbright : Look for bright pixels if true
! finddead   : Look for dead pixels if true
! withcal    : CAL_getBadPixelList is called if true
! withdead   : Dead pixels from the CAL are written to output list if true
! xbad       : array of RAWX coordinates
! ybad       : array of RAWY coordinates
! yext       : array of RAWY extensions
! tbad       : array of bad pixel types
! nbad       : number of bad pixels

    type(ArrayT),                      intent(in)    :: ev_map, med_map
    real,                              intent(in)    :: maxproba, maxratio
    logical,                           intent(in)    :: withcal, withdead,  &
                                                 findcols, findbright, finddead
    integer(kind=int16), dimension(:), intent(inout) :: xbad, ybad, yext, tbad
    integer,                           intent(inout) :: nbad

    integer(kind=int32), dimension(:,:), pointer     :: map, medsmooth
    logical,             dimension(:,:), allocatable :: goodPixel
    real,                dimension(width+2*w2,width+2*w2) :: psfmap
    real,                dimension(width+2*w2)       :: psfpro
    real    :: cpsfmed
    integer :: nx, ny, ibad, x0, y0

! Read original and smoothed arrays

    map => int32Array2Data(ev_map)
    medsmooth => int32Array2Data(med_map)

    nx = size(map,1)
    ny = size(map,2)
    if (size(medsmooth,1) /= nx .or. size(medsmooth,2) /= ny) then
       write(errstr,*) "The median-smoothed image does not have the same ",  &
                       "dimensions", size(medsmooth,1), size(medsmooth,2),  &
                       " as the original image", nx, ny
       call error("findAllBad01")
    endif

! Initialise the goodPixel array

    allocate(goodPixel(nx,ny))
    goodPixel = .true.

! Set dead and uplinked pixels to 0

    if (withcal) then
       call getCalBadpix(map,withdead,goodPixel,xbad,ybad,yext,tbad,nbad)
       call getPSF(2*w2,psfmap,psfpro)
    else
       psfmap = 1.
       psfpro = 1.
    endif

! Do not consider already known bad pixels
! This must be after getCalBadpix for the check of uplinked bright pixels

    do ibad = 1, nbad
       x0 = xbad(ibad)
       y0 = ybad(ibad)
       goodPixel(x0,y0:y0+yext(ibad)-1) = .false.
    enddo

! Set all pixels outside window to bad
! Recompute median along borders if window mode
! medsmooth runs from 0 to nx-1

    if (wx0 > 1) then
       goodPixel(1:wx0-1,:) = .false.
       medsmooth(wx0-1:wx0+w2-2,:) = spread(medsmooth(wx0+w2-1,:),1,w2)
    endif
    if (wx1 < nx) then
       goodPixel(wx1+1:nx,:) = .false.
       medsmooth(wx1-w2:wx1-1,:) = spread(medsmooth(wx1-w2-1,:),1,w2)
    endif
    if (wy0 > 1) then
       goodPixel(:,1:wy0-1) = .false.
       medsmooth(:,wy0-1:wy0+w2-2) = spread(medsmooth(:,wy0+w2-1),2,w2)
    endif
    if (wy1 < ny) then
       goodPixel(:,wy1+1:ny) = .false.
       medsmooth(:,wy1-w2:wy1-1) = spread(medsmooth(:,wy1-w2),2,w2)
    endif

! Look for dark columns, rows and pixels

    if (finddead) then

! Look for very bright pixels/columns/rows to avoid
! disrupting dark columns/rows search

       call findBadPix(map, medsmooth, goodPixel, psfmap, maxproba**2,  &
                       maxratio, .false., xbad, ybad, yext, tbad, nbad)
       if (findcols) then
          call findBadCol(map, goodPixel, psfpro, maxproba**2,  &
                          maxratio, .false., xbad, ybad, yext, tbad, nbad)
          call findBadRow(map, goodPixel, psfpro, maxproba**2,  &
                          maxratio, .false., xbad, ybad, yext, tbad, nbad)
          call findBadCol(map, goodPixel, psfpro, maxproba,  &
                          maxratio, .true., xbad, ybad, yext, tbad, nbad)
          call findBadRow(map, goodPixel, psfpro, maxproba,  &
                          maxratio, .true., xbad, ybad, yext, tbad, nbad)
       endif
       call findBadPix(map, medsmooth, goodPixel, psfmap, maxproba,  &
                       maxratio, .true., xbad, ybad, yext, tbad, nbad)
    endif

! Look for bright pixels, columns, then rows

    if (findbright) then
       call findBadPix(map, medsmooth, goodPixel, psfmap, maxproba,  &
                       maxratio, .false., xbad, ybad, yext, tbad, nbad)
       if (findcols) then
          cpsfmed = psfpro(width) / medloc(psfpro(width-w2:width+w2))
          write(errstr,*) "1-D PSF correction on median :", cpsfmed
          call message(AppMsg,Sparse)

          call findBadCol(map, goodPixel, psfpro, maxproba,  &
                          maxratio, .false., xbad, ybad, yext, tbad, nbad)

          call findBadRow(map, goodPixel, psfpro, maxproba,  &
                          maxratio, .false., xbad, ybad, yext, tbad, nbad)
       endif
    else     ! Remove very bright pixels from list
       nbad = count(tbad == DEAD)
       xbad(1:nbad) = pack(xbad,tbad == DEAD)
       ybad(1:nbad) = pack(ybad,tbad == DEAD)
       yext(1:nbad) = pack(yext,tbad == DEAD)
       tbad(1:nbad) = DEAD
    endif

    deallocate(goodPixel)

  end subroutine findAllBad

! Find the bad pixels sticking out from the noise

  subroutine findBadPix(map, medsmooth, goodPixel, psfmap, maxproba,  &
                        maxratio, negative, xbad, ybad, yext, tbad, nbad)

! map       : Image array
! medsmooth : Same image after median smoothing
! goodPixel : False for bad pixels
! psfmap    : Map of PSF around maximum
! maxproba  : Probability threshold for false detection
! maxratio  : Relative detection threshold for dark features
! negative  : Look for dead pixels instead if true
! xbad      : array of RAWX coordinates
! ybad      : array of RAWY coordinates
! yext      : array of RAWY extensions
! tbad      : array of bad pixel types
! nbad      : number of bad pixels

    integer(kind=int32), dimension(:,:), pointer       :: map, medsmooth
    logical,             dimension(:,:), intent(inout) :: goodPixel
    real, dimension(width+2*w2,width+2*w2), intent(in) :: psfmap
    real,                                intent(in)    :: maxproba, maxratio
    logical,                             intent(in)    :: negative
    integer(kind=int16), dimension(:),   intent(inout) :: xbad,ybad,yext,tbad
    integer,                             intent(inout) :: nbad

    real,                dimension(:,:), allocatable :: medroot, badtest
    logical,             dimension(width,width)      :: goodpsf
    integer,             dimension(width*width)      :: packSpace
    integer, dimension(2)   :: maxpos
    real    :: mu, mu1, mu2, mu3, mu4, cpsfmed, cpsf2
    integer :: nx, ny, thresh, ncur, ncur2, nbad0
    integer :: ibad, i0, j0, i1, j1, i, j
    logical :: source

! map and medsmooth are offset by 1 (starts at 0) with respect to other arrays

    nx = size(map,1)
    ny = size(map,2)

    allocate(medroot(nx,ny),badtest(nx,ny))

    if (negative) then

       call message(AppMsg,Verbose,"Looking for dark pixels...")

! Get security against "grey" rows or columns

       cpsfmed = maxratio
       source = .false.
    else

       call message(AppMsg,Verbose,"Looking for bright pixels...")

! Get PSF correction between median and peak
! Pack requires a mask

       goodpsf = .true.
       cpsfmed = psfmap(width,width) / medloc(pack(  &
                psfmap(width-w2:width+w2,width-w2:width+w2),goodpsf))
       write(errstr,*) "2-D PSF correction on median :", cpsfmed
       call message(AppMsg,Sparse)
    endif

! Estimate the square root of the local counts
! Systematically add 1 to median to avoid false detections at low counts

    medroot = sqrt(real(medsmooth+1))

! Estimate the local excesses in rms units

    badtest = (map - medsmooth*cpsfmed) / medroot

    nbad0 = nbad
    do ibad = 1,size(xbad)-nbad

! Look for maximum local excess where goodPixel is True
! Should not be used with pointers

       if (negative) then
          maxpos = minloc(badtest,goodPixel)
       else
          maxpos = maxloc(badtest,goodPixel)
       endif

! Stop if less than 3 sigmas

       if (.not. negative .and. badtest(maxpos(1),maxpos(2)) <  3.)  exit
       if (      negative .and. badtest(maxpos(1),maxpos(2)) > -3.)  exit

       goodPixel(maxpos(1),maxpos(2)) = .false.
       write(errstr,*) "Investigating", maxpos, badtest(maxpos(1),maxpos(2))
       call message(AppMsg,Noisy)

! Define width x width reference zone

       i0 = min(max(maxpos(1),wx0+w2),wx1-w2)
       j0 = min(max(maxpos(2),wy0+w2),wy1-w2)
       ncur = count(goodPixel(i0-w2:i0+w2,j0-w2:j0+w2))
       if (ncur < 5) then
          write(errstr,*) "Less than 5 good pixels around", maxpos,  &
                          ". Counted as bad."
          call warning("findBadPix11")
       else

! Get median local value outside suspect pixel, add 1
! Correct for PSF profile

          i = width + i0 - maxpos(1)
          j = width + j0 - maxpos(2)

          if (negative) then
             mu = medloc(pack(map(i0-w2-1:i0+w2-1,j0-w2-1:j0+w2-1),  &
                               goodPixel(i0-w2:i0+w2,j0-w2:j0+w2),packSpace))
             thresh = quantile(mu,-maxproba/(ncur+1))
          else
             cpsfmed = psfmap(width,width) / medloc(pack(  &
     psfmap(i-w2:i+w2,j-w2:j+w2),goodPixel(i0-w2:i0+w2,j0-w2:j0+w2)))
             mu1 = medloc(pack(map(i0-w2-1:i0+w2-1,j0-w2-1:j0+w2-1),  &
                               goodPixel(i0-w2:i0+w2,j0-w2:j0+w2),packSpace)) &
                 * cpsfmed + 1

! Get average local value outside suspect pixel, add 1 for security
! (better than median when very small)

             mu2 = (sum(map(i0-w2-1:i0+w2-1,j0-w2-1:j0+w2-1),  &
                        goodPixel(i0-w2:i0+w2,j0-w2:j0+w2))+1)  &
                 * psfmap(width,width)  &
                 / sum(psfmap(i-w2:i+w2,j-w2:j+w2),  &
                       goodPixel(i0-w2:i0+w2,j0-w2:j0+w2))
             mu = min(mu1,mu2)
             source = .true.

! Get median and average in smaller box (better for PSF correction)

             i1 = min(max(maxpos(1),wx0+w2-1),wx1-w2+1)
             j1 = min(max(maxpos(2),wy0+w2-1),wy1-w2+1)
             ncur2 = count(goodPixel(i1-w2+1:i1+w2-1,j1-w2+1:j1+w2-1))
             if (ncur2 >= 5) then
                i = width + i1 - maxpos(1)
                j = width + j1 - maxpos(2)
                cpsf2 = psfmap(width,width) / medloc(pack(  &
                         psfmap(i-w2+1:i+w2-1,j-w2+1:j+w2-1),  &
                         goodPixel(i1-w2+1:i1+w2-1,j1-w2+1:j1+w2-1)))
                mu3 = medloc(pack(map(i1-w2:i1+w2-2,j1-w2:j1+w2-2),  &
                       goodPixel(i1-w2+1:i1+w2-1,j1-w2+1:j1+w2-1),packSpace)) &
                    * cpsf2 + 1
                mu4 = (sum(map(i1-w2:i1+w2-2,j1-w2:j1+w2-2),  &
                           goodPixel(i1-w2+1:i1+w2-1,j1-w2+1:j1+w2-1))+1)  &
                    * psfmap(width,width)  &
                    / sum(psfmap(i-w2+1:i+w2-1,j-w2+1:j+w2-1),  &
                          goodPixel(i1-w2+1:i1+w2-1,j1-w2+1:j1+w2-1))

! Take minimum of all the estimates
! Set source to False only if inner corona is lower than outer (conservative)

                if (min(mu3,mu4) < mu+1.e-5) then
                   mu = min(mu3,mu4)
                   source = (mu3-1)/cpsf2 > (mu1-1)/cpsfmed
                endif
             endif

! Get minimum forbidden value at required probability threshold

             thresh = quantile(mu,maxproba/(ncur+1))
          endif

          write(errstr,*) "Pixel value:", map(maxpos(1)-1,maxpos(2)-1),  &
                          "   Local average:", mu, "   Threshold:", thresh,  &
                          "   Source:", source

          if ((negative .and. map(maxpos(1)-1,maxpos(2)-1)>thresh) .or.  &
              (.not.negative .and. map(maxpos(1)-1,maxpos(2)-1)<thresh)) then

! Continue search if the bad pixel was a source

             if (source) then
                call message(AppMsg,Noisy)
                goodPixel(maxpos(1),maxpos(2)) = .true.
                badtest(maxpos(1),maxpos(2)) = 0.
                cycle
             else
                call message(AppMsg,Verbose)
                exit
             endif
          else
             call message(AppMsg,Noisy)
             do j = max(maxpos(2)-w2,wy0),min(maxpos(2)+w2,wy1)
                j0 = min(max(j,wy0+w2),wy1-w2)
                j1 = width + j0 - j
                do i = max(maxpos(1)-w2,wx0),min(maxpos(1)+w2,wx1)
                   i0 = min(max(i,wx0+w2),wx1-w2)
                   i1 = width + i0 - i
                   if (medroot(i,j) > 1.1) then

! Recompute local median + 1
! map and medsmooth start at 0

                      medsmooth(i-1,j-1) = medloc(pack(  &
                                 map(i0-w2-1:i0+w2-1,j0-w2-1:j0+w2-1),  &
                                 goodPixel(i0-w2:i0+w2,j0-w2:j0+w2),packSpace))
                      medroot(i,j) = sqrt(real(medsmooth(i-1,j-1)+1))
                      cpsfmed = psfmap(width,width) / medloc(pack(  &
                                 psfmap(i1-w2:i1+w2,j1-w2:j1+w2),  &
                                 goodPixel(i0-w2:i0+w2,j0-w2:j0+w2)))
                      badtest(i,j) = (map(i-1,j-1) -  &
                                     medsmooth(i-1,j-1)*cpsfmed) / medroot(i,j)
                   endif
                enddo
             enddo
          endif
       endif
       nbad = nbad + 1
       xbad(nbad) = maxpos(1)
       ybad(nbad) = maxpos(2)
       yext(nbad) = 1
       if (negative) then
          tbad(nbad) = DEAD
       else
          tbad(nbad) = HOT
       endif
    enddo

    goodPixel(maxpos(1),maxpos(2)) = .true.

    if (negative) then
       write(errstr,*) nbad-nbad0, " dark pixels found in image"
    else
       write(errstr,*) nbad-nbad0, " bright pixels found in image"
    endif
    call message(AppMsg,Sparse)

    deallocate(medroot,badtest)

  end subroutine findBadPix

! Find bad columns and bad segments within

  subroutine findBadCol(map, goodPixel, psfpro, maxproba, maxratio, &
                        negative, xbad, ybad, yext, tbad, nbad)

! map       : Image array
! goodPixel : False for bad pixels
! psfpro    : PSF projected on 1 dimension (around maximum)
! maxproba  : Probability threshold for false detection
! maxratio  : Relative detection threshold for dark features
! negative  : Look for dead pixels instead if true
! xbad      : array of RAWX coordinates
! ybad      : array of RAWY coordinates
! yext      : array of RAWY extensions
! tbad      : array of bad pixel types
! nbad      : number of bad pixels

    integer(kind=int32), dimension(:,:), pointer       :: map
    logical,             dimension(:,:), intent(inout) :: goodPixel
    real,         dimension(width+2*w2), intent(in)    :: psfpro
    real,                                intent(in)    :: maxproba, maxratio
    logical,                             intent(in)    :: negative
    integer(kind=int16), dimension(:),   intent(inout) :: xbad,ybad,yext,tbad
    integer,                             intent(inout) :: nbad

    integer,             dimension(:),   allocatable :: profil, npixels
    integer,             dimension(1024)             :: cbad, ubad, xseg, lseg
    real,                dimension(1024)             :: pixrate
    integer :: nx, nbc, nseg
    integer :: ic, is

    if (negative) then
       call message(AppMsg,Verbose,"Looking for dark columns...")
    else
       call message(AppMsg,Verbose,"Looking for bright columns...")
    endif

! Project the image onto RAWX and look for bad columns

    nx = size(map,1)
    allocate(profil(nx),npixels(nx))
    profil  = sum(map,2,goodPixel)
    npixels = count(goodPixel,2)

    call findBad1D(profil, npixels, psfpro, maxproba, maxratio, negative,  &
                   pixrate, cbad, ubad, nbc)

    if (negative) then
       write(errstr,*) nbc, " dark columns found in image"
    else
       write(errstr,*) nbc, " bright columns found in image"
    endif
    call message(AppMsg,Sparse)

! Check whether only part of the column is bright

    do ic = 1, nbc
       write(errstr,*) "Column", cbad(ic)
       call message(AppMsg,Noisy)
       call findBadSeg(map(cbad(ic)-1,:), goodPixel(cbad(ic),:),  &
                       pixrate(ic), negative, xseg, lseg, nseg)
       do is = 1, nseg
          nbad = nbad + 1
          xbad(nbad) = cbad(ic)
          ybad(nbad) = xseg(is)
          yext(nbad) = lseg(is)
          tbad(nbad) = ubad(ic)
          goodPixel(cbad(ic),xseg(is):xseg(is)+lseg(is)-1) = .false.
       enddo
    enddo

    deallocate(profil,npixels)

  end subroutine findBadCol

! Find bad rows and bad segments within

  subroutine findBadRow(map, goodPixel, psfpro, maxproba, maxratio, &
                        negative, xbad, ybad, yext, tbad, nbad)

! map       : Image array
! goodPixel : False for bad pixels
! psfpro    : PSF projected on 1 dimension (around maximum)
! maxproba  : Probability threshold for false detection
! maxratio  : Relative detection threshold for dark features
! negative  : Look for dead pixels instead if true
! xbad      : array of RAWX coordinates
! ybad      : array of RAWY coordinates
! yext      : array of RAWY extensions
! tbad      : array of bad pixel types
! nbad      : number of bad pixels

    integer(kind=int32), dimension(:,:), pointer       :: map
    logical,             dimension(:,:), intent(inout) :: goodPixel
    real,         dimension(width+2*w2), intent(in)    :: psfpro
    real,                                intent(in)    :: maxproba, maxratio
    logical,                             intent(in)    :: negative
    integer(kind=int16), dimension(:),   intent(inout) :: xbad,ybad,yext,tbad
    integer,                             intent(inout) :: nbad

    integer,             dimension(:),   allocatable :: profil, npixels
    integer,             dimension(1024)             :: cbad, ubad, xseg, lseg
    real,                dimension(1024)             :: pixrate
    integer :: ny, nbc, nseg
    integer :: ic, is, i

    if (negative) then
       call message(AppMsg,Verbose,"Looking for dark rows...")
    else
       call message(AppMsg,Verbose,"Looking for bright rows...")
    endif

! Must exchange X and Y window parameters

    i = wx0
    wx0 = wy0
    wy0 = i
    i = wx1
    wx1 = wy1
    wy1 = i

! Project the image onto RAWY and look for bad rows

    ny = size(map,2)
    allocate(profil(ny),npixels(ny))
    profil  = sum(map,1,goodPixel)
    npixels = count(goodPixel,1)

    call findBad1D(profil, npixels, psfpro, maxproba, maxratio, negative,  &
                   pixrate, cbad, ubad, nbc)

    if (negative) then
       write(errstr,*) nbc, " dark rows found in image"
    else
       write(errstr,*) nbc, " bright rows found in image"
    endif
    call message(AppMsg,Sparse)

! Check whether only part of the row is bright

    do ic = 1, nbc
       write(errstr,*) "Row", cbad(ic)
       call message(AppMsg,Noisy)
       call findBadSeg(map(:,cbad(ic)-1), goodPixel(:,cbad(ic)),  &
                       pixrate(ic), negative, xseg, lseg, nseg)
       do is = 1, nseg
          do i = 0, lseg(is)-1
             nbad = nbad + 1
             xbad(nbad) = xseg(is) + i
             ybad(nbad) = cbad(ic)
             yext(nbad) = 1
             tbad(nbad) = ubad(ic)
          enddo
          goodPixel(xseg(is):xseg(is)+lseg(is)-1,cbad(ic)) = .false.
       enddo
    enddo

    deallocate(profil,npixels)

! Reset X and Y window parameters

    i = wx0
    wx0 = wy0
    wy0 = i
    i = wx1
    wx1 = wy1
    wy1 = i

  end subroutine findBadRow

! Find statistical peaks in a 1-D distribution

  subroutine findBad1D(profil, npixels, psfpro, maxproba, maxratio, negative, &
                       pixrate, cbad, tbad, nbc)

! profil   : Number of counts in each bin
! npixels  : Depth of each bin
! psfpro   : PSF profile
! maxproba : Probability threshold for false detection
! maxratio : Relative detection threshold for dark features
! negative : Look for dead pixels instead if true
! pixrate  : Upper limit for pixel acceptability
! cbad     : List of peaks identified
! tbad     : List of bad pixel types (HOT or DEAD)
! nbc      : Number of peaks identified

    integer, dimension(:),    intent(in)  :: profil, npixels
    real,    dimension(width+2*w2), intent(in) :: psfpro
    real,                     intent(in)  :: maxproba, maxratio
    logical,                  intent(in)  :: negative
    real,    dimension(1024), intent(out) :: pixrate
    integer, dimension(1024), intent(out) :: cbad, tbad
    integer,                  intent(out) :: nbc

    real,    dimension(:),    allocatable :: medsmooth,medroot,ratecol,badtest
    logical, dimension(:),    allocatable :: goodCol
    integer, dimension(1)                 :: maxpos
    real    :: colmu, colmu2, colthresh, cpsfmed, cpsf2, colmed
    integer :: i, i0, i1, ibad, ncur, ncur2, nx
    logical :: source

    nbc = 0

    if (wx1-wx0+1 < width) then
       write(errstr,*) "Width for local analysis", width,  &
                    " is larger than window size", wx1-wx0+1, ". Nothing done."
       call warning("findBad1D10")
       return
    endif

! Get the average number of counts per pixel

    nx = size(profil)
    allocate(medsmooth(nx), medroot(nx), ratecol(nx), badtest(nx), goodCol(nx))
    goodCol = npixels > 0
    where (goodCol)
       ratecol = real(profil) / npixels
    endwhere

! Get the median over width pixels. First iteration

    do i = wx0, wx1
       medsmooth(i) = getMedian(ratecol, goodCol, width, wx0, wx1, i,  &
                                ncur, cbad, nbc) * npixels(i)
    enddo

    if (negative) then

! Get security against "grey" rows or columns

       cpsfmed = maxratio
       source = .false.
    else

! Get PSF correction factor

       cpsfmed = psfpro(width) / medloc(psfpro(width-w2:width+w2))
    endif

! Estimate the square root of the local counts
! Systematically add 1 to median to avoid false detections at low counts

    medroot = sqrt(medsmooth+1.)

! Estimate the local excesses in rms units

    badtest = (profil - medsmooth*cpsfmed) / medroot

    do ibad = 1,nx

! Look for maximum local excess where goodCol is True
! Should not be used with pointers

       if (negative) then
          maxpos = minloc(badtest,goodCol)
       else
          maxpos = maxloc(badtest,goodCol)
       endif

! Stop if less than 3 sigmas

       if (.not. negative .and. badtest(maxpos(1)) <  3.)  exit
       if (      negative .and. badtest(maxpos(1)) > -3.)  exit

       goodCol(maxpos(1)) = .false.
       write(errstr,*) "Investigating", maxpos, badtest(maxpos(1))
       call message(AppMsg,Noisy)

! Compute local median excluding the suspect pixel

       pixrate(nbc+1) = getMedian(ratecol, goodCol, width, wx0, wx1,  &
                                  maxpos(1), ncur, cbad, nbc)

       if (ncur < 2) then
          write(errstr,*) "Less than 2 good rows/columns around", maxpos(1),  &
                          ". Counted as bad."
          call warning("findBad1D11")
       else

! Multiply by number of good pixels

          i1 = min(max(maxpos(1),wx0+w2),wx1-w2)
          i = width + i1 - maxpos(1)
          if (.not. negative)  cpsfmed = psfpro(width)  &
                                       / medloc(pack(psfpro(i-w2:i+w2),  &
                                           goodCol(i1-w2:i1+w2)))
          colmu = pixrate(nbc+1) * npixels(maxpos(1)) * cpsfmed

          if (negative) then
             colthresh = quantile(max(colmu-sqrt(colmu/ncur)-1.,0.),  &
                                  -maxproba/(ncur+1))
          else

! Get median and in smaller box (better for PSF correction)

             source = .true.
             colmed = getMedian(ratecol, goodCol, width-2, wx0, wx1,  &
                                maxpos(1), ncur2, cbad, nbc)
             if (ncur2 >= 2) then
                i1 = min(max(maxpos(1),wx0+w2-1),wx1-w2+1)
                i = width + i1 - maxpos(1)
                cpsf2 = psfpro(width) / medloc(pack(psfpro(i-w2+1:i+w2-1),  &
                                           goodCol(i1-w2+1:i1+w2-1)))
                colmu2 = colmed * npixels(maxpos(1)) * cpsf2

! Take minimum of both estimates
! Set source to False only if inner value is lower than outer (conservative)

                if (colmu2+sqrt(colmu2/ncur2) < colmu+sqrt(colmu/ncur)) then
                   colmu = colmu2
                   source = colmu2/cpsf2 > colmu/cpsfmed
                else
                   ncur2 = ncur
                endif
             else
                ncur2 = ncur
             endif

! Get maximum allowed value at required probability threshold
! Variance of mu is about mu/ncur, add 1 to be conservative

             colthresh = quantile(colmu+sqrt(colmu/ncur2)+1.,maxproba/(ncur+1))
          endif

          write(errstr,*) "Pixel value:", profil(maxpos(1)),  &
                       "   Local median:", colmu, "   Threshold:", colthresh, &
                       "   Source:", source

          if ((negative .and. profil(maxpos(1))>colthresh) .or.  &
              (.not.negative .and. profil(maxpos(1))<colthresh)) then

! Continue search if the bad pixel was a source

             if (source) then
                call message(AppMsg,Noisy)
                goodCol(maxpos(1)) = .true.
                badtest(maxpos(1)) = 0.
                cycle
             else
                call message(AppMsg,Verbose)
                exit
             endif
          else
             call message(AppMsg,Noisy)

! Recompute local median + 1

             do i = max(maxpos(1)-w2,wx0),min(maxpos(1)+w2,wx1)
                medsmooth(i) = npixels(i)  &
                     * getMedian(ratecol,goodCol,width,wx0,wx1,i,ncur,cbad,nbc)
                medroot(i) = sqrt(medsmooth(i)+1)
                if (goodCol(i)) then
                   i0 = min(max(i,wx0+w2),wx1-w2)
                   i1 = width + i0 - i
                   if (.not. negative)  cpsfmed = psfpro(width)  &
                                          / medloc(pack(psfpro(i1-w2:i1+w2),  &
                                              goodCol(i0-w2:i0+w2)))
                   badtest(i) = (profil(i) - medsmooth(i)*cpsfmed) / medroot(i)
                else
                   badtest(i) = 0.
                endif
             enddo
          endif
       endif
       nbc = nbc + 1
       cbad(nbc) = maxpos(1)
       if (negative) then
          tbad(nbc) = DEAD
       else
          tbad(nbc) = HOT
       endif
    enddo

    deallocate(medsmooth, medroot, ratecol, badtest, goodCol)

  end subroutine findBad1D

! Find bright segments within a bright column

  subroutine findBadSeg(profil, goodPixel, pixrate, negative, xseg, lseg, nseg)

! profil    : Number of counts in each pixel
! goodPixel : False if pixel is already known to be bad
! pixrate   : Expected pixel rate
! negative  : Look for dark segments within a dark column if true
! xseg      : List of first pixels in segments
! lseg      : List of segment lengths
! nseg      : Number of segments identified

    integer, dimension(:),    intent(in)  :: profil
    logical, dimension(:),    intent(in)  :: goodPixel
    real,                     intent(in)  :: pixrate
    logical,                  intent(in)  :: negative
    integer, dimension(1024), intent(out) :: xseg, lseg
    integer,                  intent(out) :: nseg

    integer, dimension(:),    allocatable :: boxcounts
    logical, dimension(:),    allocatable :: goodcur
    integer, dimension(1)                 :: maxpos

! Confidence level for stopping accumulating bad segments

    real, parameter  :: conflevel = 0.1

    integer :: wseg, ws2, nctmax, nx, nctmin, nleft, nloc, cleft, zero
    integer :: i, i0, i1, j0, j1, is

! Define minimum width to have at least 1 expected count

    if (negative) then
       cleft = sum(profil,goodPixel)
       nleft = count(goodPixel)
       wseg = int(nleft/real(cleft)) + 1
    else
       wseg = int(1./pixrate) + 1
    endif
    ws2 = max(wseg/2,1)
    wseg = 2 * ws2 + 1

! Integrate with boxcar window

    nx = size(profil)
    allocate(goodcur(nx),boxcounts(nx))

! Define value at which pixels will never be found bad

    if (negative) then
       zero = cleft*wseg + 1
    else
       zero = 0
    endif

    goodcur = goodPixel
    boxcounts = zero
    do i = wy0+ws2,wy1-ws2
       if (any(goodPixel(i-ws2:i+ws2)))  &
          boxcounts(i) = sum(profil(i-ws2:i+ws2),goodPixel(i-ws2:i+ws2))  &
                       * wseg / real(count(goodPixel(i-ws2:i+ws2)))
    enddo

! Pick segments starting with maximum value

    nseg = 0
    do is = 1, nx

       nleft = count(goodcur)
       if (nleft == 0)  exit

! Find position of maximum and maximum number of counts acceptable

       if (negative) then
          maxpos = minloc(boxcounts)
          nctmax = quantile(pixrate*nleft,-conflevel)
       else
          maxpos = maxloc(boxcounts)
          nctmax = quantile(pixrate*nleft,conflevel)
       endif
       i0 = maxpos(1) - ws2
       i1 = maxpos(1) + ws2

! If nctmax = 0 then reject whole column

       cleft = sum(profil,goodcur)
       if ((nctmax==0 .and. .not.negative) .or. (cleft==0 .and. negative)) then
          nseg = 1
          xseg(nseg) = 1
          lseg(nseg) = nx
          goodcur = .false.
          exit
       endif

! Test whether remaining number of counts is reasonable
! Exit if local count rate is reasonable as well or no counts left

       if (negative) then
          if (cleft > nctmax) then
             nctmin = quantile(pixrate*count(goodcur(i0:i1)),  &
                               -wseg*conflevel/nleft)
             nloc = sum(profil(i0:i1),goodcur(i0:i1))
             write(errstr,*) "Counts in segment:", nloc, "   Minimum:", nctmin
             call message(AppMsg,Noisy)
             if (nloc > nctmin) then
                if (nloc == 0 .and. nseg > 0) then

! Exit only if not adjacent to rejected area
! While nloc=0 the segments accumulate to the right

                   if (i0 > xseg(nseg)+lseg(nseg))  exit
                else
                   exit
                endif
             endif
          endif
       else if (cleft < nctmax) then
          if (cleft == 0) then
             exit
          else
             nctmin = quantile(pixrate*count(goodcur(i0:i1)),  &
                               wseg*conflevel/nleft)
             nloc = sum(profil(i0:i1),goodcur(i0:i1))
             write(errstr,*) "Counts in segment:", nloc, "   Maximum:", nctmin
             call message(AppMsg,Noisy)
             if (nloc < nctmin)  exit
          endif
       endif

! Extend to near border or other segment if less than wseg good pixels
! Some bad pixels may be isolated (found as individual bad pixels)
! Segments may overlap

       if (i0 >= wy0+wseg) then
          if (.not. all(goodcur(i0-wseg:i0-1))) then
             do i = i0-1,i0-wseg,-1
                if (.not. goodcur(i))  exit
             enddo
             i0 = i + 1
          endif
       endif
       if (i0 < wy0+wseg)  i0 = wy0
       if (i1 <= wy1-wseg) then
          if (.not. all(goodcur(i1+1:i1+wseg))) then
             do i = i1+1,i1+wseg
                if (.not. goodcur(i))  exit
             enddo
             i1 = i - 1
          endif
       endif
       if (i1 > wy1-wseg)  i1 = wy1
       nseg = nseg + 1
       xseg(nseg) = i0
       lseg(nseg) = i1 - i0 + 1

! Set new segment to bad

       goodcur(i0:i1) = .false.

! Recompute boxcounts, so that next segment is found as far as possible

       boxcounts(i0:i1) = zero
       do j0 = max(i0-ws2-1,wy0+ws2),i0-1
          if ( goodcur(j0) )  exit
       enddo
       if (j0 < i0) then
          boxcounts(j0+1:i0-1) = zero
          boxcounts(j0) = sum(profil(j0-ws2:j0+ws2),goodcur(j0-ws2:j0+ws2))  &
                        * wseg / real(count(goodcur(j0-ws2:j0+ws2)))
       endif
       do j1 = min(i1+ws2+1,wy1-ws2),i1+1,-1
          if ( goodcur(j1) )  exit
       enddo
       if (j1 > i1) then
          boxcounts(i1+1:j1-1) = zero
          boxcounts(j1) = sum(profil(j1-ws2:j1+ws2),goodcur(j1-ws2:j1+ws2))  &
                        * wseg / real(count(goodcur(j1-ws2:j1+ws2)))
       endif
    enddo

    if (nleft == 0) then
       write(errstr,*) "Found entirely bad."
    else if (negative) then
       write(errstr,*) "Identifying segments of width", wseg, " up to", &
                       nctmax, " counts left."
    else
       write(errstr,*) "Identifying segments of width", wseg, " down to", &
                       nctmax, " counts left."
    endif
    call message(AppMsg,Noisy)

    deallocate(goodcur,boxcounts)

  end subroutine findBadSeg

! Get local median in 1-D array

  real function getMedian(rateCol,goodCol,length,wx0,wx1,i,ncur,cbad,nbc)

! rateCol : input array
! goodCol : list of indices to consider
! length   : length for median averaging
! wx0,wx1 : Bounds within which the median should be taken
! i       : current index
! ncur    : number of good pixels in window
! cbad    : List of peaks identified
! nbc     : Number of peaks identified

    real,    dimension(:),    intent(in)    :: rateCol
    logical, dimension(:),    intent(inout) :: goodCol
    integer,                  intent(in)    :: length, wx0, wx1, i
    integer,                  intent(out)   :: ncur
    integer, dimension(1024), intent(inout) :: cbad
    integer,                  intent(inout) :: nbc


    real,    dimension(:),    allocatable   :: rate, dist
    integer, dimension(1)                   :: posmin, posmax
    real    :: aver, grad, iav
    integer :: i0, k, imin, imax, pmin, pmax

    i0 = min(max(i-length/2,wx0),wx1-length+1)
    ncur = count(goodCol(i0:i0+length-1))

    allocate(rate(length),dist(length))
    rate = rateCol(i0:i0+length-1)

! Remove underlying gradient if large statistics (more than 1 ct/pixel)
! Do it only if first and last pixels are extrema

    if (minval(rate,goodCol(i0:i0+length-1)) > 1 .and. ncur > 1) then
       dist = (/ (real(k), k=i0,i0+length-1) /)
       imin = minval(dist,goodCol(i0:i0+length-1))
       posmin = minloc(rate,goodCol(i0:i0+length-1)) + i0 - 1
       imax = maxval(dist,goodCol(i0:i0+length-1))
       posmax = maxloc(rate,goodCol(i0:i0+length-1)) + i0 - 1
       pmin = min(posmin(1),posmax(1))
       pmax = max(posmin(1),posmax(1))
       if (pmin == imin .and. pmax == imax) then
          aver = sum(rate,goodCol(i0:i0+length-1)) / ncur
          iav = sum(dist,goodCol(i0:i0+length-1)) / ncur
          dist = dist - iav
          grad = sum((rate-aver)*dist,goodCol(i0:i0+length-1)) /  &
                 sum(dist**2,goodCol(i0:i0+length-1))
          rate = rate - grad * dist
       endif
    endif

    if (goodCol(i)) then
       if (ncur < 3) then
          getMedian = 0.
          write(errstr,*) "Less than 2 good rows/columns around", i,  &
                          ". Counted as bad."
          call warning("getMedian11")
          nbc = nbc + 1
          cbad(nbc) = i
          goodCol(i) = .false.
       else
          getMedian = medloc(pack(rate,goodCol(i0:i0+length-1)))
       endif
    else if (ncur < 2) then
       getMedian = 0.
    else
       getMedian = medloc(pack(rate,goodCol(i0:i0+length-1)))
    endif

    deallocate(rate,dist)

  end function getMedian

! Compute quantiles for the Poisson distribution
! Probability to get quantile or above is always <= epsilon
! gaussCvf must be a separate F77 routine to be accessible by fzero
! Get lower quantile if epsilon < 0

  integer function quantile(mu, epsilon)

! mu      : average value
! epsilon : probability level

    real, intent(in) :: mu, epsilon

    real, external :: gaussCvf
    real(double)   :: target, mud, cur, cumul
    real           :: b, c, r
    integer        :: ier
    logical        :: negative

    negative = epsilon < 0.

    if (mu >= 30) then

! Initialize gaussCvf

       b = gaussCvf(-abs(epsilon))

! Use Gauss limit for Poisson distribution
! Find lower and upper limit to threshold

       if (gaussCvf(1.) > 0) then
          b = max(sqrt(-log(3.*abs(epsilon))/2.),1.)
          c = sqrt(-log(2.*abs(epsilon))*2.)
       else
          b = 0.
          c = 1.
       endif
       r = c
       call fzero(gaussCvf, b, c, r, 1.e-5, 1.e-5, ier)
       if (ier >= 3) then
          write(errstr,*) "error in call to fzero: iflag =", ier
          call error("quantile01")
       endif
       if (negative) then
          quantile = int(mu - sqrt(mu)*b)
       else
          quantile = int(mu + sqrt(mu)*b) + 1
       endif
    else

! Compute cumulative distribution iteratively

       mud = real(mu,kind=double)
       if (negative) then
          target = real(-epsilon,kind=double) * exp(mud)
       else
          target = (1.d0 - real(epsilon,kind=double)) * exp(mud)
       endif
       quantile = 0
       cur = 1.d0
       cumul = cur
       do while (cumul <= target)
          quantile = quantile + 1
          cur = cur * mud / quantile
          cumul = cumul + cur
       enddo
       if (negative)  quantile = quantile - 1
    endif

  end function quantile

! Compute median of an array

  integer function intmedloc(toto)

! toto: 1-D input array of integer values

    integer, dimension(:), intent(in)  :: toto

    integer, dimension(:), allocatable :: tata, tutu
    integer :: nelem

    nelem = size(toto)
    allocate(tata(nelem),tutu(nelem))
    tutu = toto
    call isort(tutu,tata,nelem,1)
    intmedloc = (tutu((nelem+1)/2) + tutu(nelem/2+1)) / 2
    deallocate(tata,tutu)

  end function intmedloc

! Compute median of an array

  real function realmedloc(toto)

! toto: 1-D input array of real values

    real,    dimension(:), intent(in)  :: toto

    real,    dimension(:), allocatable :: tata, tutu
    integer :: nelem

    nelem = size(toto)
    allocate(tata(nelem),tutu(nelem))
    tutu = toto
    call ssort(tutu,tata,nelem,1)
    realmedloc = 0.5 * (tutu((nelem+1)/2) + tutu(nelem/2+1))
    deallocate(tata,tutu)

  end function realmedloc

! Compute Y extent of bad pixels

  subroutine mergeBad(xbad, ybad, yext, tbad, fbad, nbad)

! xbad     : array of RAWX coordinates
! ybad     : array of RAWY coordinates
! yext     : array of RAWY extensions
! tbad     : array of bad pixel types
! fbad     : array of bad pixel status (uplinked or not)
! nbad     : number of bad pixels

    integer(kind=int16), dimension(:), intent(inout) :: xbad, ybad, yext,  &
                                                        tbad, fbad
    integer,                           intent(inout) :: nbad

    integer, dimension(:), allocatable :: xbad2, ybad2, yext2, tbad2, fbad2,  &
                                          perm
    integer :: nbad2, ier, k1, k2, m1, m2, ymin, ymax

    if (nbad <= 1)  return

    allocate(xbad2(nbad), ybad2(nbad), yext2(nbad), tbad2(nbad), fbad2(nbad))
    allocate(perm(nbad))

! Sort bad pixels per column

    ymin = minval(ybad)
    ymax = maxval(ybad+yext-1)
    ybad2 = (xbad(1:nbad)-1)*(ymax-ymin+1) + ybad(1:nbad)-ymin
    call ipsort(ybad2, nbad, perm, 1, ier)
    if (ier > 0) then
       write(errstr,*) "error in call to ipsort : ier =", ier
       call error("mergeBad01")
    endif

! Loop on bad pixels

    k1 = 1
    do nbad2 = 1, nbad
       m1 = perm(k1)
       ymax = ybad(m1) + yext(m1)

! Look for all bad pixels vertically adjacent to that one

       do k2 = k1+1, nbad
          m2 = perm(k2)
          if (xbad(m2) > xbad(m1) .or. tbad(m2) /= tbad(m1)  &
                                  .or. fbad(m2) /= fbad(m1))  exit
          if (ybad(m2) > ymax)  exit
          ymax = max(ymax,int(ybad(m2)+yext(m2)))
       enddo
       xbad2(nbad2) = xbad(m1)
       ybad2(nbad2) = ybad(m1)
       yext2(nbad2) = ymax - ybad(m1)
       tbad2(nbad2) = tbad(m1)
       fbad2(nbad2) = fbad(m1)
       if (k2 > nbad)  exit
       k1 = k2
    enddo

! Replace routine arguments

    nbad = nbad2
    xbad(1:nbad) = xbad2(1:nbad)
    ybad(1:nbad) = ybad2(1:nbad)
    yext(1:nbad) = yext2(1:nbad)
    tbad(1:nbad) = tbad2(1:nbad)
    fbad(1:nbad) = fbad2(1:nbad)
    deallocate(xbad2, ybad2, yext2, tbad2, fbad2, perm)

  end subroutine mergeBad

! Get the uploaded bright pixels and the dead pixels from the CCF
! goodPixel should be all True where map > 0 on input

  subroutine getCalBadpix(map,withdead,goodPixel,xbad,ybad,yext,tbad,nbad)

! map       : Image array
! goodPixel : array of good pixels
! withdead : Dead pixels are written to output list if true
! xcal     : array of RAWX coordinates
! ycal     : array of RAWY coordinates
! yext     : array of RAWY extensions
! tbad     : array of bad pixel types
! nbad     : number of bad pixels

    integer(kind=int32), dimension(:,:), pointer       :: map
    logical,                             intent(in)    :: withdead
    logical,             dimension(:,:), intent(inout) :: goodPixel
    integer(kind=int16), dimension(:),   intent(inout) :: xbad,ybad,yext,tbad
    integer,                             intent(inout) :: nbad

    integer(kind=int16), dimension(:), pointer :: rawX, rawY, type, yextent
    logical,             dimension(:), pointer :: uplinked
    integer(kind=int16), dimension(:), pointer :: fcal
    integer :: ibad, xcal, ycal, ymax, yhot, locsum, xloc, yloc, ncal

    call CAL_getBadPixelList(rawX, rawY, type, yextent, uplinked)
    ncal = size(rawX)

! Order and group the CAL bad pixels

    allocate(fcal(0:ncal-1))
    fcal = 1
    call mergeBad(rawX, rawY, yextent, type, fcal, ncal)

    do ibad = 0, ncal-1

! Use standard integers for indices

       xcal = rawX(ibad)
       ycal = rawY(ibad)

! Ignore bad pixels outside current window
! Set them not uplinked to avoid failures in following loop

       if (ycal < wy0) then
          yextent(ibad) = yextent(ibad) - wy0 + ycal
          ycal = wy0
       endif
       if (xcal<wx0 .or. xcal>wx1 .or. ycal>wy1 .or. yextent(ibad)<=0) then
          uplinked(ibad) = .false.
          cycle
       endif
       yextent(ibad) = min(yextent(ibad),int(wy1-ycal+1,kind=int16))

       if (uplinked(ibad))  goodPixel(xcal,ycal:ycal+yextent(ibad)-1) = .false.

! Optionally write dead pixels to list

       if (withdead .and. type(ibad)==DEAD) then
          nbad = nbad + 1
          xbad(nbad) = xcal
          ybad(nbad) = ycal
          yext(nbad) = yextent(ibad)
          tbad(nbad) = type(ibad)
       endif
    enddo

! Check uplinked bad pixels are absent from the data (only check maximum)
! 2nd loop allows to know whether neighbouring pixels are declared bad as well

    do ibad = 0, ncal-1
       if (uplinked(ibad)) then
          xcal = rawX(ibad)
          ycal = rawY(ibad)
          ymax = ycal
          do yhot = ycal+1, ycal+yextent(ibad)-1
             if (map(xcal-1,ymax-1) < map(xcal-1,yhot-1))  ymax = yhot
          enddo

! Clear error only when maximum > sum(neighbours) because of pattern projection

          locsum = 0
          do xloc = max(xcal-1,wx0), min(xcal+1,wx1)
             if (goodPixel(xloc,ymax))  locsum = locsum + map(xloc-1,ymax-1)
          enddo
          do yloc = max(ymax-1,wy0), min(ymax+1,wy1)
             if (goodPixel(xcal,yloc))  locsum = locsum + map(xcal-1,yloc-1)
          enddo

          if (map(xcal-1,ymax-1) > locsum) then
             write(errstr,*) "Bright pixel at", xcal, ymax,  &
                " is declared uplinked in the CCF but is present in the data."
             call warning("getCalBadpix10")

! Reactivate search for the whole column segment

             goodPixel(xcal,ycal:ycal+yextent(ibad)-1) = .true.
          endif
       endif
    enddo

    call CAL_releaseMemory(rawX, rawY)
    call CAL_releaseMemory(type, yextent)
    call CAL_releaseMemory(uplinked)

    if (withdead .and. nbad > 0) then
       write(errstr,*) "Included", nbad, " dead pixels in list"
       call message(AppMsg, Verbose)
       write(errstr,*) "Xdead:", xbad(1:nbad)
       call message(AppMsg, Noisy)
       write(errstr,*) "Ydead:", ybad(1:nbad)
       call message(AppMsg, Noisy)
    endif

  end subroutine getCalBadpix

! Get the point Spread Function near the maximum from the CAL

  subroutine getPSF(wp2,psfmap,psfpro)

! wp2    : half-width of box
! psfmap : PSF in 2*wp2+1 x 2*wp2+1 area around maximum
! psfpro : PSF projected on X axis in 2*wp2+1 around maximum

    integer,                             intent(in)  :: wp2
    real, dimension(2*wp2+1,2*wp2+1),    intent(out) :: psfmap
    real, dimension(2*wp2+1),            intent(out) :: psfpro

    integer(kind=int16), dimension(2),   parameter   :: xin = (/1,2/),  &
                                                        yin = (/1,1/)
    real(double),        dimension(:),   pointer     :: xmilli, ymilli, zmilli
    real(single),        dimension(:,:), pointer     :: image
    integer,             dimension(2)                :: maxpos2
    integer,             dimension(1)                :: maxpos1
    type(PsfT)        :: psf
    type(PsfBinSizeT) :: psfbin
    type(PsfWindowT)  :: psfwindow
    real(double)      :: pixsize
    integer           :: accuracy, nx, ny, nx2, ny2
    logical           :: random

! Get former accuracy level and randomization

    call CAL_getState(accuracyLevel=accuracy, randomize=random)

! Set accuracy to high (multi-gaussian representation) and no randomization

    call CAL_setState(accuracyLevel=ACCURACY_HIGH, randomize=.false.)

! Get pixel size in mm

    call CAL_rawXY2mm(xin, yin, xmilli, ymilli, zmilli,  &
                      DETECTOR_OPTICSCENTER_SYSTEM)
    pixsize = sqrt((xmilli(1)-xmilli(0))**2+(ymilli(1)-ymilli(0))**2)
    psfbin%binX = pixsize
    psfbin%binY = pixsize
    call CAL_releaseMemory(xmilli, ymilli, zmilli)

! Get handle to PSF at center of FOV

    psf = CAL_getPSF(1.d3, 0.d0, 0.d0)

! Set bin size to one pixel

    call CAL_psfSetBinSize(psf, psfbin)

! Get PSF in slightly larger window along X, 600 along Y

    psfwindow%halfX = (wp2+2) * pixsize
    psfwindow%halfY = (wy1-wy0+1)/2 * pixsize
    call CAL_psfImage(psf, psfwindow, image)
    nx = size(image,1)
    ny = size(image,2)

! Find location of maximum (not always at center)
! maxloc should not be used with pointers

    nx2 = (nx-1) / 2
    ny2 = (ny-1) / 2
    psfmap = image(nx2-wp2:nx2+wp2,ny2-wp2:ny2+wp2)
    maxpos2 = maxloc(psfmap)

! image starts at 0

    maxpos2(1) = min(max(maxpos2(1)+nx2-wp2-1,wp2),nx-wp2-1)
    maxpos2(2) = min(max(maxpos2(2)+ny2-wp2-1,wp2),ny-wp2-1)
    psfmap = image(maxpos2(1)-wp2:maxpos2(1)+wp2,maxpos2(2)-wp2:maxpos2(2)+wp2)

! Project on X axis
! Location of maximum is not the same for MOS2

    psfpro = sum(image(maxpos2(1)-wp2:maxpos2(1)+wp2,:),2)
    maxpos1 = maxloc(psfpro) - 1
    if (maxpos1(1) /= maxpos2(1)) then
       maxpos1(1) = min(max(maxpos1(1)+maxpos2(1)-wp2,wp2),nx-wp2-1)
       psfpro = sum(image(maxpos1(1)-wp2:maxpos1(1)+wp2,:),2)
    endif

    call CAL_releaseMemory(image)

! Get back to former accuracy level and randomization

    call CAL_setState(accuracyLevel=accuracy, randomize=random)

! Fix to avoid false bad pixels at source core in MOS1

    call fixPsf(psfmap,psfpro)

  end subroutine getPsf

! Hardcoded version of the PSF for MOS1, taken from 0087_0124930101

  subroutine fixPsf(psfmap,psfpro)

! psfmap : PSF in 9 x 9 area around maximum
! psfpro : PSF projected on X axis in 9 pixels around maximum

    real, dimension(4*w2+1,4*w2+1),    intent(inout) :: psfmap
    real, dimension(4*w2+1),           intent(inout) :: psfpro

    integer, dimension(4*w2+1,4*w2+1), parameter     :: psf1 = reshape( &
         (/ 2172, 2401, 3042, 3432, 4014, 3758, 3789, 3561, 2593,  &
            2802, 3256, 3703, 4970, 5764, 5379, 5143, 4528, 2727,  &
            3248, 4232, 5232, 7399, 8762, 8225, 6633, 5019, 3077,  &
            3338, 4620, 7156,10567,12366,11303, 7430, 4963, 3686,  &
            3575, 5496, 8768,12318,15549,14510, 8938, 5331, 4023,  &
            3554, 4983, 7115, 9632,11472,12892, 9160, 4928, 3587,  &
            3350, 4024, 4843, 6494, 7191, 8054, 7032, 4610, 3353,  &
            2381, 2776, 3138, 3739, 4812, 5206, 4973, 3754, 2729,  &
            1606, 1917, 2275, 2527, 3436, 3821, 4106, 3079, 2306 /),  &
         shape=(/4*w2+1,4*w2+1/))
    integer, dimension(4*w2+1),        parameter     :: psfx =  &
         (/ 51645, 64819, 84476, 97324,100262, 85109, 66143, 52354, 43637 /)
    integer, dimension(4*w2+1),        parameter     :: psfy =  &
         (/ 45094, 55929, 72173, 87628,104472, 95108, 74553, 54865, 43963 /)

    logical, dimension(width,width) :: goodpsf
    integer :: instrument
    real    :: cpsfmed

    call CAL_getState(instrument=instrument)
    if (instrument /= EMOS1)  return

! Fix of 2-D PSF

    goodpsf = .true.
    cpsfmed = psfmap(width,width) / medloc(pack(  &
              psfmap(width-w2:width+w2,width-w2:width+w2),goodpsf))
    if (cpsfmed < 1.6) then
       call message(AppMsg, Sparse, "2-D PSF replaced by hardcoded version")
       psfmap = psf1
    endif

! Fix of 1-D PSF
! Use projection onto Y (more peaked ==> more conservative)

    cpsfmed = psfpro(width) / medloc(psfpro(width-w2:width+w2))
    if (cpsfmed < 1.1) then
       call message(AppMsg, Sparse, "1-D PSF replaced by hardcoded version")
       psfpro = psfy
    endif

  end subroutine fixPsf

! Write bad pixels file

  subroutine writeBadpix(bad_tab, xbad, ybad, yext, tbad, fbad, nbad)

! bad_tab : Handle to the output bad pixels table
! xbad    : array of RAWX coordinates
! ybad    : array of RAWY coordinates
! yext    : array of RAWY extensions
! tbad    : array of bad pixel types
! fbad    : array of bad pixel status (uplinked or not)
! nbad    : number of bad pixels

    type(TableT),                      intent(in) :: bad_tab
    integer(kind=int16), dimension(:), intent(in) :: xbad, ybad, yext,  &
                                                     tbad, fbad
    integer,                           intent(in) :: nbad

    type(ColumnT)  :: badx_col, bady_col, badt_col, bade_col, badf_col
    integer(kind=int16), dimension(:), pointer :: badx, bady, badtype,  &
                                                  badext, badflag
    integer :: nrows

! Open output bad pixels file

    write(errstr,*) "Writing", nbad, " rows to ", trim(qualifiedName(bad_tab))
    call message(AppMsg,Sparse)

    nrows = numberOfRows(bad_tab)
    if (nbad > nrows)  call insertRows(bad_tab, nrows, nbad-nrows)

    badx_col = column(bad_tab, "RAWX", Modify)
    bady_col = column(bad_tab, "RAWY", Modify)
    badt_col = column(bad_tab, "TYPE", Modify)
    bade_col = column(bad_tab, "YEXTENT", Modify)
    badf_col = column(bad_tab, "BADFLAG", Modify)

    badx    => int16Data(badx_col)
    bady    => int16Data(bady_col)
    badtype => int16Data(badt_col)
    badext  => int16Data(bade_col)
    badflag => int16Data(badf_col)

    badtype = tbad(1:nbad)
    badflag = fbad(1:nbad)
    badext  = yext(1:nbad)
    badx    = xbad(1:nbad)
    bady    = ybad(1:nbad)

    call release(parent(bad_tab))

  end subroutine writeBadpix

! Write bad pixels file

  subroutine readBadpix(bad_tab, incremental,  &
                        xbad, ybad, yext, tbad, fbad, nbad)

! bad_tab : Handle to the bad pixels table
! incremental : Add new bad pixels to bad pixels file if true
! xbad    : array of RAWX coordinates
! ybad    : array of RAWY coordinates
! yext    : array of RAWY extensions
! tbad    : array of bad pixel types
! fbad    : array of bad pixel status (uplinked or not)
! nbad    : number of bad pixels

    type(TableT),                      intent(in)  :: bad_tab
    logical,                           intent(in)  :: incremental
    integer(kind=int16), dimension(:), intent(out) :: xbad, ybad, yext,  &
                                                      tbad, fbad
    integer,                           intent(out) :: nbad

    type(ColumnT)  :: badx_col, bady_col, badt_col, bade_col, badf_col
    integer(kind=int16), dimension(:), pointer :: badx, bady, badtype,  &
                                                  badext, badflag

    if (incremental) then

       nbad = numberOfRows(bad_tab)
       write(errstr,*) "Found", nbad, " bad pixels in ",  &
                       trim(qualifiedName(bad_tab))
       call message(AppMsg,Sparse)

       badx_col = column(bad_tab, "RAWX", Modify)
       bady_col = column(bad_tab, "RAWY", Modify)
       badt_col = column(bad_tab, "TYPE", Modify)
       bade_col = column(bad_tab, "YEXTENT", Modify)
       badf_col = column(bad_tab, "BADFLAG", Modify)

       badx    => int16Data(badx_col)
       bady    => int16Data(bady_col)
       badtype => int16Data(badt_col)
       badext  => int16Data(bade_col)
       badflag => int16Data(badf_col)

       tbad(1:nbad) = badtype
       fbad(1:nbad) = badflag
       yext(1:nbad) = badext
       xbad(1:nbad) = badx
       ybad(1:nbad) = bady

    else
       badx_col = addColumn(bad_tab, "RAWX", Integer16)
       bady_col = addColumn(bad_tab, "RAWY", Integer16)
       badt_col = addColumn(bad_tab, "TYPE", Integer16)
       bade_col = addColumn(bad_tab, "YEXTENT", Integer16)
       badf_col = addColumn(bad_tab, "BADFLAG", Integer16)
       nbad = 0
    endif

    call release(badx_col)
    call release(bady_col)
    call release(badt_col)
    call release(bade_col)
    call release(badf_col)

  end subroutine readBadpix

! Parameter reading

  subroutine getParamValues(ev_map, med_map, bad_tab, incremental, maxproba,  &
                   maxratio, findcols, findbright, finddead, withcal, withdead)

! ev_map      : Handle to the input image
! med_map     : Handle to the median smoothed image
! bad_tab     : Handle to the output bad pixels table
! incremental : Add new bad pixels to bad pixels file if true
! maxproba    : Probability threshold for false detection
! maxratio    : Relative detection threshold for dark features
! findcols    : Look for bad rows or columns if true
! findbright  : Look for bright pixels if true
! finddead    : Look for dead pixels if true
! withcal     : CAL_getBadPixelList is called if true
! withdead    : Dead pixels are written to output list if true

    type(ArrayT),        intent(out) :: ev_map, med_map
    type(TableT),        intent(out) :: bad_tab
    real,                intent(out) :: maxproba, maxratio
    logical,             intent(out) :: withcal, withdead, findcols,  &
                                        findbright, finddead, incremental

    type(DataSetT) :: im_set, med_set, bad_set
    integer, dimension(:), pointer :: sizarray
    character(len=1024) :: image_out, median_out, bad_out
    integer :: itest

! Open input image file

    image_out = stringParameter(outimage)
    write(errstr,*) "Opening image file : ",trim(image_out)
    call message(AppMsg,Sparse)
    im_set = dataSet(image_out, Read)
    ev_map = array(im_set, "PRIMARY")
    sizarray => dimensions(ev_map)

! Get window parameters

    if (hasAttribute(ev_map,"WINDOWX0")) then
       wx0 = int32Attribute(ev_map,"WINDOWX0")
       wx1 = wx0 + int32Attribute(ev_map,"WINDOWDX") - 1
       wy0 = int32Attribute(ev_map,"WINDOWY0")
       wy1 = wy0 + int32Attribute(ev_map,"WINDOWDY") - 1
       if (wx0 > 1) then
          write(errstr,*) "Window limits:", wx0, wx1, wy0, wy1
          call message(AppMsg,Sparse)
       endif
       wx0 = max(1,wx0)
       wx1 = min(sizarray(0),wx1)
       wy0 = max(1,wy0)
       wy1 = min(sizarray(1),wy1)
    else
       wx0 = 1
       wx1 = sizarray(0)
       wy0 = 1
       wy1 = sizarray(1)
    endif

! Open smoothed image file

    median_out = stringParameter(medimage)
    write(errstr,*) "Opening smoothed image file : ",trim(median_out)
    call message(AppMsg,Sparse)
    med_set = dataSet(median_out, Read)
    med_map = array(med_set, "PRIMARY")

! Check compatibility of keywords between map and smoothed map

    if (.not. equalKeywords(attributable(ev_map), &
                            attributable(med_map),keycompat,.false.) .or.  &
        .not. equalKeywords(attributable(im_set), &
                            attributable(med_set),keycompat,.false.)) then
       write(errstr,*) "The smoothed image ", trim(median_out),  &
          " does not correspond to the original image ", trim(image_out),  &
          ". Some keywords differ among",  &
          (" ",keycompat(itest),itest=1,size(keycompat))
       call error("getParamValues04")
    endif

! Threshold

    maxproba = realParameter(threshold)
    write(errstr,*) "False detection probability  :",maxproba," per pixel"
    call message(AppMsg,Sparse)

! Columns

    findcols = booleanParameter(badcols)
    if (findcols)  call message(AppMsg,Sparse,  &
       "The task will also look for bad segments in rows or columns")

! Bright pixels

    findbright = booleanParameter(onbright)
    if (findbright)  call message(AppMsg,Sparse,  &
       "The task will look for bright pixels")

! Dead pixels

    finddead = booleanParameter(ondead)
    if (finddead) then
       call message(AppMsg,Sparse,"The task will look for dark pixels")
       maxratio = realParameter(maxrat)
       write(errstr,*) "The task will look for dark pixels less than",  &
                       maxratio, " relative to their neighbours"
       call message(AppMsg,Sparse)
    endif

! Output file name

    bad_out   = stringParameter(outbadpix)
    incremental = booleanParameter(incbadpix)
    if (incremental .and. .not.setExists(bad_out)) then
       write(errstr,*) trim(bad_out), " does not exist. ",  &
                       "incremental mode ignored."
       call warning("getParamValues10")
       incremental = .false.
    endif

    if (incremental) then
       bad_set = dataSet(bad_out, Modify)
       bad_tab = table(bad_set, "BADPIX")

! Check compatibility of keywords between map and bad pixels file
! Ignore OBS_ID and EXPIDSTR

       if (.not. equalKeywords(attributable(ev_map),attributable(bad_tab), &
                               keycompat(1:4),.false.) .or.  &
           .not. equalKeywords(attributable(im_set),attributable(bad_set), &
                               keycompat(1:4),.false.)) then
          write(errstr,*) "The bad pixels file ", trim(bad_out),  &
             " does not correspond to the image ", trim(image_out),  &
             ". Some keywords differ among",(" ",keycompat(itest),itest=1,4)
          call error("getParamValues03")
       endif
       write(errstr,*) "Bad pixels will be read from and written to ",  &
                       trim(bad_out)
       call message(AppMsg,Sparse)
    else
       write(errstr,*) "Bad pixels will be written to ",trim(bad_out)
       call message(AppMsg,Sparse)
       bad_set = dataSet(bad_out, Create)
       call setAttributes(attributable(bad_set),attributable(im_set))
       bad_tab = addTable(bad_set, "BADPIX", 0)
       call setAttributes(attributable(bad_tab), attributable(ev_map))
    endif

! Set CAL state

    withcal = booleanParameter(usecal)

    if (withcal) then
       call CAL_setState(parent(ev_map))
       withdead = booleanParameter(adddead)
       if (withdead)  call message(AppMsg,Sparse,  &
          "Dead pixels from the CAL will be included in the bad pixels list")
    endif

  end subroutine getParamValues

end module embadpixfind_mod
